"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WebSocketClient = void 0;
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const WebSocket = require("ws");
const common_1 = require("../../common");
const socket_1 = require("../../socket");
const url_util_1 = require("../../common/util/url.util");
class WebSocketClient {
    _info;
    _store;
    _options;
    _client;
    constructor(options, store) {
        this._info = {
            type: options.type,
            address: url_util_1.UrlUtil.removeSearchParams(options.address),
            originalAddress: options.address,
            connected: false,
            connecting: false,
        };
        this._options = options;
        this._store = store;
    }
    get store() {
        return this._store;
    }
    async connect() {
        if (!this._info.connected) {
            this.clear();
            this._client = new WebSocket(this._info.originalAddress, {
                rejectUnauthorized: false,
                headers: this._options.headers,
            });
            this.onMessage(this._client);
            this.onError(this._client);
            this.onClose(this._client);
            try {
                const open$ = (0, rxjs_1.fromEvent)(this._client, 'open').pipe(common_1.RxJSUtil.timeout(this._options.connectionTimeout, `connection to ${this._info.address} is timed out`), (0, operators_1.tap)(() => this._info.connected = true), (0, operators_1.tap)(() => this._store.add({
                    type: socket_1.SocketEventType.Connected,
                    date: new Date(),
                    message: this._info,
                })), (0, operators_1.map)(() => undefined));
                this._info.connecting = true;
                await (0, rxjs_1.firstValueFrom)(open$);
                this._info.connecting = false;
                this._info.connected = true;
            }
            catch (error) {
                this.clear();
                throw error;
            }
        }
    }
    sendData(data) {
        const requestIdInfo = common_1.MessageUtil.findRequestId(data);
        const packed = common_1.MessageUtil.packToString(data);
        const eventMessage = {
            format: packed.format,
            data: packed.data
        };
        if (typeof data === 'object' && data !== null) {
            eventMessage.data = data;
        }
        return this.send(packed.data, eventMessage, requestIdInfo);
    }
    async sendBytes(bytes) {
        const encoding = this._options.binaryEncoding;
        const packed = common_1.MessageUtil.packToBuffer(bytes);
        const eventMessage = {
            format: packed.format,
            size: bytes.length,
            data: bytes
        };
        if (encoding) {
            eventMessage[encoding] = packed.data.toString(encoding);
        }
        return this.send(packed.data, eventMessage);
    }
    close() {
        this._client?.terminate();
        this.clear();
    }
    clear() {
        this._client?.removeAllListeners();
        this._client = undefined;
        this._info.connected = false;
        this._info.connecting = false;
    }
    onMessage(client) {
        client.on('message', (message, isBinary) => {
            const unpacked = isBinary
                ? common_1.MessageUtil.unpack(message)
                : common_1.MessageUtil.unpack(message.toString());
            const encoding = this._options.binaryEncoding;
            let eventMessage = unpacked;
            if (unpacked.format === common_1.MessageFormat.ByteArray) {
                eventMessage = {
                    format: unpacked.format,
                    size: unpacked.data.length,
                    data: unpacked.data,
                };
                if (encoding) {
                    eventMessage[encoding] = (Buffer.from(unpacked.data)).toString(encoding);
                }
            }
            this._store.add({
                type: socket_1.SocketEventType.ReceivedMessage,
                date: new Date(),
                message: eventMessage,
            });
        });
    }
    onError(client) {
        client.on('error', err => this._store.add({
            type: socket_1.SocketEventType.Error,
            date: new Date(),
            message: {
                message: err.message
            },
        }));
    }
    onClose(client) {
        client.on('close', (code, reason) => {
            if (this._info.connected) {
                this._store.add({
                    type: socket_1.SocketEventType.Closed,
                    date: new Date(),
                    message: { code, reason: reason.toString() },
                });
                this.clear();
                this.retryConnection();
            }
        });
    }
    retryConnection() {
        const retryInterval = setInterval(async () => {
            try {
                if (this._info.connected) {
                    clearInterval(retryInterval);
                }
                else if (!this._info.connecting) {
                    this._store.add({
                        type: socket_1.SocketEventType.Reconnecting,
                        date: new Date(),
                        message: this._info,
                    });
                    await this.connect();
                    clearInterval(retryInterval);
                }
            }
            catch (err) { }
        }, this._options.reconnectionInterval);
    }
    send(data, eventMessage, requestIdInfo) {
        const client = this._client;
        const connected = this._info.connected;
        if (!client || !connected) {
            throw new Error(`error: client is not connected to ${this._info.address}`);
        }
        const currentDate = new Date();
        return new Promise(async (resolve, reject) => {
            client.send(data, { binary: Buffer.isBuffer(data) }, async (err) => {
                if (!err) {
                    const event = {
                        type: socket_1.SocketEventType.SentMessage,
                        date: new Date(),
                        message: eventMessage,
                    };
                    const messageId = this._store.add(event);
                    const reply = requestIdInfo
                        ? await this.awaitReply(requestIdInfo, currentDate)
                        : undefined;
                    resolve({ messageId, reply });
                }
                reject(err);
            });
        });
    }
    async awaitReply(requestIdInfo, awaitAfterDate) {
        const isRecord = (event) => {
            return typeof event.message.data === 'object'
                && event.message.data !== null
                && !Array.isArray(event.message.data);
        };
        const isReply = (event, info) => {
            return event.message.data[info.key] === info.value;
        };
        const reply$ = this._store.listen().pipe((0, operators_1.filter)(event => event.type === socket_1.SocketEventType.ReceivedMessage), (0, operators_1.filter)(event => event.date >= awaitAfterDate), (0, operators_1.filter)(event => isRecord(event)), (0, operators_1.filter)(event => isReply(event, requestIdInfo)), (0, operators_1.map)(event => event.message), (0, operators_1.timeout)(this._options.replyTimeout));
        let reply;
        try {
            reply = await (0, rxjs_1.firstValueFrom)(reply$);
        }
        catch (err) { }
        return reply;
    }
}
exports.WebSocketClient = WebSocketClient;
